==> ppcboot-0.5.1.tar.gz: ppcboot-0.5.1/common/cmd_nvedit.c <==
/*
 * (C) Copyright 2000
 * Wolfgang Denk, DENX Software Engineering, wd@denx.de.
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

/*
 * Support for persistent environment data
 */
#include <ppcboot.h>
#include <command.h>
#include <cmd_nvedit.h>

/*
 * The environment storages is simply a list of '\0'-terminated
 * "name=value" strings, the end of the list marked by a double '\0'.
 * New entries are always addrd at the end. Deleting an entry shifts
 * the remaining entries to the front. Replacing an entry is a
 * combination of deleting the old and adding the new value.
 */

extern uchar environment[];
extern ulong env_size;

#if (CONFIG_COMMANDS & CFG_CMD_ENV)
static uchar *flash_addr = environment;
#endif

static uchar *envmatch (uchar *, uchar *);
static void env_init(void);

#define XMK_STR(x)	#x
#define MK_STR(x)	XMK_STR(x)

static uchar default_environment[] = {
	"bootargs="	CONFIG_BOOTARGS			"\0"
	"bootcmd="	CONFIG_BOOTCOMMAND		"\0"
#if (CONFIG_BOOTDELAY >= 0)
	"bootdelay="	MK_STR(CONFIG_BOOTDELAY)	"\0"
#endif
#if (CONFIG_BAUDRATE >= 0)
	"baudrate="	MK_STR(CONFIG_BAUDRATE)		"\0"
#endif
#ifdef	CONFIG_ETHADDR
	"ethaddr="	MK_STR(CONFIG_ETHADDR)		"\0"
#endif
#ifdef	CONFIG_IPADDR
	"ipaddr="	MK_STR(CONFIG_IPADDR)		"\0"
#endif
#ifdef	CONFIG_SERVERIP
	"serverip="	MK_STR(CONFIG_SERVERIP)		"\0"
#endif
	"\0"
};

char *getenv (uchar *name)
{
	uchar *env, *nxt;

	env_init();

	for (env=environment; *env; env=nxt+1) {
		char *val;

		for (nxt=env; *nxt; ++nxt)
			;
		val=envmatch(name, env);
		if (!val)
			continue;
		return (val);
	}

	return (NULL);
}

void do_printenv (cmd_tbl_t *cmdtp, bd_t *bd, int flag, int argc, char *argv[])
{
	uchar *env, *nxt;
	int i;

	env_init();

	if (argc == 1) {		/* Print all env variables	*/
		for (env=environment; *env; env=nxt+1) {
			for (nxt=env; *nxt; ++nxt)
				;
			serial_putstr (env);
			serial_putc  ('\n');
		}
		return;
	}

	for (i=1; i<argc; ++i) {	/* print single env variables	*/
		char *name = argv[i];
		char *val = NULL;

		for (env=environment; *env; env=nxt+1) {

			for (nxt=env; *nxt; ++nxt)
				;
			val=envmatch(name, env);
			if (val) {
				serial_putstr (name);
				serial_putc   ('=');
				serial_putstr (val);
				serial_putc  ('\n');
				break;
			}
		}
		if (!val)
			printf ("## Error: \"%s\" not defined\n", name);
	}
}

void do_setenv   (cmd_tbl_t *cmdtp, bd_t *bd, int flag, int argc, char *argv[])
{
	int  i, len;
	uchar *env, *nxt;
	uchar *oldval = NULL;
	uchar *name;

	env_init();

	if (argc < 2) {
		printf ("Usage:\n%s\n", cmdtp->usage);
		return;
	}

	name = argv[1];

	/*
	 * search if variable with this name already exists
	 */
	for (env=environment; *env; env=nxt+1) {
		for (nxt=env; *nxt; ++nxt)
			;
		if ((oldval=envmatch(name, env)) != NULL)
			break;
	}

	/*
	 * Delete any existing definition
	 */
	if (oldval) {
		/*
		 * Ethernet Address and serial# can be set only once
		 */
		if ((strcmp (name, "ethaddr") == 0) ||
		    (strcmp (name, "serial#") == 0) ) {
			printf ("Can't overwrite \"%s\"\n", name);
			return;
		}

		if (*++nxt == '\0') {
			*env = '\0';
		} else {
			for (;;) {
				*env = *nxt++;
				if ((*env == '\0') && (*nxt == '\0'))
					break;
				++env;
			}
		}
		*++env = '\0';
	}

	/* Delete only ? */
	if (argc < 3)
		return;

	/*
	 * Append new definition at the end
	 */
	for (env=environment; *env || *(env+1); ++env)
		;
	if (env > environment)
		++env;
	/*
	 * Overflow when:
	 * "name" + "=" + "val" +"\0\0"  > env_size - (env-environment)
	 */
	len = strlen(name) + 2;
	/* add '=' for first arg, ' ' for all others */
	for (i=2; i<argc; ++i) {
		len += strlen(argv[i]) + 1;
	}
	if (len > (&environment[env_size]-env)) {
		printf ("## Error: environment overflow, \"%s\" deleted\n", name);
		return;
	}
	while ((*env = *name++) != '\0')
		env++;
	for (i=2; i<argc; ++i) {
		char *val = argv[i];

		*env = (i==2) ? '=' : ' ';
		while ((*++env = *val++) != '\0')
			;
	}

	/* end is marked with double '\0' */
	*++env = '\0';

        /* Changes of the Ethernet or IP address should be reflected
         * in the board info structure.
	 */

	if (strcmp(argv[1],"ethaddr") == 0) {
		char *s = argv[2];	/* always use only one arg */
		char *e;
		for (i=0; i<6; ++i) {
			bd->bi_enetaddr[i] = s ? simple_strtoul(s, &e, 16) : 0;
			if (s) s = (*e) ? e+1 : e;
		}
	} else if (strcmp(argv[1],"ipaddr") == 0) {
		char *s = argv[2];	/* always use only one arg */
		char *e;
		bd->bi_ip_addr = 0;
		for (i=0; i<4; ++i) {
			ulong val = s ? simple_strtoul(s, &e, 10) : 0;
			bd->bi_ip_addr <<= 8;
			bd->bi_ip_addr  |= (val & 0xFF);
			if (s) s = (*e) ? e+1 : e;
		}
	}
}

#if (CONFIG_COMMANDS & CFG_CMD_ENV)

void do_saveenv  (cmd_tbl_t *cmdtp, bd_t *bd, int flag, int argc, char *argv[])
{
	int rc;
	extern void flash_sect_protect (int p, ulong addr_first, ulong addr_last);
	extern void flash_sect_erase (ulong addr_first, ulong addr_last);
#ifdef CONFIG_CPCI405
        ulong sector_flash_addr;
        ulong sector_flash_size;
        ulong sector_flash_offs;
        int i;
        flash_info_t *info;

	env_init();

        /*
         * Calculate environment variables sector address and size
         */
        info = addr2info((ulong)flash_addr);
        for (i=0; i<info->sector_count; i++)
          {
            if (info->start[i] >= (ulong)flash_addr)
              break;
          }
        sector_flash_addr = info->start[i-1];
        sector_flash_size = info->start[i] - info->start[i-1];
        sector_flash_offs = (ulong)flash_addr - info->start[i-1];

#if 1
        printf("\nflash_addr = %08lx sector_addr=%08lx size=%08lx offs=%08lx\n",
               (ulong)flash_addr, sector_flash_addr, sector_flash_size, sector_flash_offs);
#endif

        /*
         * Copy sector down to ram
         */
	memcpy((uchar *)0x10000, (uchar *)sector_flash_addr, sector_flash_size);

        /*
         * Copy new environment variables to ram image of flash sector
         */
	memcpy((uchar *)(0x10000+sector_flash_offs), (uchar *)environment, env_size);

	flash_sect_protect (0, sector_flash_addr, sector_flash_addr+sector_flash_size-1);

	printf ("Erasing Flash...");
	flash_sect_erase (sector_flash_addr, sector_flash_addr+sector_flash_size-1);

	printf ("Saving Environment to Flash...\n");
	switch (rc = flash_write ((uchar *)0x10000, sector_flash_addr, sector_flash_size)) {
	case 0: break;
	case 1: printf ("Timeout writing to Flash\n");
		break;
	case 2: printf ("Flash not Erased\n");
		break;
	case 4: printf ("Can't write to protected Flash sectors\n");
		break;
	default:
		printf ("%s[%d] FIXME: rc=%d\n",__FILE__,__LINE__,rc);
	}
	
	flash_sect_protect (1, sector_flash_addr, sector_flash_addr+sector_flash_size-1);
#else	/* ! CONFIG_CPCI405 */

	env_init();

	flash_sect_protect (0, (ulong)flash_addr, (ulong)flash_addr+env_size-1);
	
	printf ("Erasing Flash...");
	flash_sect_erase ((ulong)flash_addr, (ulong)flash_addr+env_size-1);
	
	printf ("Saving Environment to Flash...\n");
	switch (rc = flash_write (environment, (ulong)flash_addr, env_size)) {
	case 0: break;
	case 1: printf ("Timeout writing to Flash\n");
		break;
	case 2: printf ("Flash not Erased\n");
		break;
	case 4: printf ("Can't write to protected Flash sectors\n");
		break;
	case 8: printf ("Outside available Flash\n");
		return;
	default:
		printf ("%s[%d] FIXME: rc=%d\n",__FILE__,__LINE__,rc);
	}
	
	flash_sect_protect (1, (ulong)flash_addr, (ulong)flash_addr+env_size-1);

#endif	/* CONFIG_CPCI405 */
}

#endif	/* CFG_CMD_ENV */

/*
 * s1 is either a simple 'name', or a 'name=value' pair.
 * s2 is a 'name=value' pair.
 * If the names match, return the value of s2, else NULL.
 */

static uchar *
envmatch (uchar *s1, uchar *s2)
{

	while (*s1 == *s2++)
		if (*s1++ == '=')
			return(s2);
	if (*s1 == '\0' && *(s2-1) == '=')
		return(s2);
	return(NULL);
}

/*
 * Prevent confusion if running from erased flash memory
 */
static void env_init(void)
{
	if (environment[0] == 0xFF) {
		printf ("*** Warning - no Environment, using defaults\n\n");
		memcpy (environment,
			default_environment,
			sizeof(default_environment));
	}
}

==> ppcboot-0.5.2.tar.gz: ppcboot-0.5.2/common/cmd_nvedit.c <==
/*
 * (C) Copyright 2000
 * Wolfgang Denk, DENX Software Engineering, wd@denx.de.
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

/*
 * Support for persistent environment data
 */
#include <ppcboot.h>
#include <command.h>
#include <cmd_nvedit.h>

/*
 * The environment storages is simply a list of '\0'-terminated
 * "name=value" strings, the end of the list marked by a double '\0'.
 * New entries are always addrd at the end. Deleting an entry shifts
 * the remaining entries to the front. Replacing an entry is a
 * combination of deleting the old and adding the new value.
 */

extern uchar environment[];
extern ulong env_size;

/* need both ENV and flash */
#if ((CONFIG_COMMANDS & (CFG_CMD_ENV|CFG_CMD_FLASH)) == (CFG_CMD_ENV|CFG_CMD_FLASH))
static uchar *flash_addr = environment;
#endif

static uchar *envmatch (uchar *, uchar *);
static void env_init(void);

#define XMK_STR(x)	#x
#define MK_STR(x)	XMK_STR(x)

static uchar default_environment[] = {
	"bootargs="	CONFIG_BOOTARGS			"\0"
	"bootcmd="	CONFIG_BOOTCOMMAND		"\0"
#if (CONFIG_BOOTDELAY >= 0)
	"bootdelay="	MK_STR(CONFIG_BOOTDELAY)	"\0"
#endif
#if (CONFIG_BAUDRATE >= 0)
	"baudrate="	MK_STR(CONFIG_BAUDRATE)		"\0"
#endif
#ifdef	CONFIG_ETHADDR
	"ethaddr="	MK_STR(CONFIG_ETHADDR)		"\0"
#endif
#ifdef	CONFIG_IPADDR
	"ipaddr="	MK_STR(CONFIG_IPADDR)		"\0"
#endif
#ifdef	CONFIG_SERVERIP
	"serverip="	MK_STR(CONFIG_SERVERIP)		"\0"
#endif
	"\0"
};

char *getenv (uchar *name)
{
	uchar *env, *nxt;

	env_init();

	for (env=environment; *env; env=nxt+1) {
		char *val;

		for (nxt=env; *nxt; ++nxt)
			;
		val=envmatch(name, env);
		if (!val)
			continue;
		return (val);
	}

	return (NULL);
}

void do_printenv (cmd_tbl_t *cmdtp, bd_t *bd, int flag, int argc, char *argv[])
{
	uchar *env, *nxt;
	int i;

	env_init();

	if (argc == 1) {		/* Print all env variables	*/
		for (env=environment; *env; env=nxt+1) {
			for (nxt=env; *nxt; ++nxt)
				;
			serial_putstr (env);
			serial_putc  ('\n');
		}
		return;
	}

	for (i=1; i<argc; ++i) {	/* print single env variables	*/
		char *name = argv[i];
		char *val = NULL;

		for (env=environment; *env; env=nxt+1) {

			for (nxt=env; *nxt; ++nxt)
				;
			val=envmatch(name, env);
			if (val) {
				serial_putstr (name);
				serial_putc   ('=');
				serial_putstr (val);
				serial_putc  ('\n');
				break;
			}
		}
		if (!val)
			printf ("## Error: \"%s\" not defined\n", name);
	}
}

void do_setenv   (cmd_tbl_t *cmdtp, bd_t *bd, int flag, int argc, char *argv[])
{
	int  i, len;
	uchar *env, *nxt;
	uchar *oldval = NULL;
	uchar *name;

	env_init();

	if (argc < 2) {
		printf ("Usage:\n%s\n", cmdtp->usage);
		return;
	}

	name = argv[1];

	/*
	 * search if variable with this name already exists
	 */
	for (env=environment; *env; env=nxt+1) {
		for (nxt=env; *nxt; ++nxt)
			;
		if ((oldval=envmatch(name, env)) != NULL)
			break;
	}

	/*
	 * Delete any existing definition
	 */
	if (oldval) {
		/*
		 * Ethernet Address and serial# can be set only once
		 */
		if ((strcmp (name, "ethaddr") == 0) ||
		    (strcmp (name, "serial#") == 0) ) {
			printf ("Can't overwrite \"%s\"\n", name);
			return;
		}

		if (*++nxt == '\0') {
			*env = '\0';
		} else {
			for (;;) {
				*env = *nxt++;
				if ((*env == '\0') && (*nxt == '\0'))
					break;
				++env;
			}
		}
		*++env = '\0';
	}

	/* Delete only ? */
	if (argc < 3)
		return;

	/*
	 * Append new definition at the end
	 */
	for (env=environment; *env || *(env+1); ++env)
		;
	if (env > environment)
		++env;
	/*
	 * Overflow when:
	 * "name" + "=" + "val" +"\0\0"  > env_size - (env-environment)
	 */
	len = strlen(name) + 2;
	/* add '=' for first arg, ' ' for all others */
	for (i=2; i<argc; ++i) {
		len += strlen(argv[i]) + 1;
	}
	if (len > (&environment[env_size]-env)) {
		printf ("## Error: environment overflow, \"%s\" deleted\n", name);
		return;
	}
	while ((*env = *name++) != '\0')
		env++;
	for (i=2; i<argc; ++i) {
		char *val = argv[i];

		*env = (i==2) ? '=' : ' ';
		while ((*++env = *val++) != '\0')
			;
	}

	/* end is marked with double '\0' */
	*++env = '\0';

        /* Changes of the Ethernet or IP address should be reflected
         * in the board info structure.
	 */

	if (strcmp(argv[1],"ethaddr") == 0) {
		char *s = argv[2];	/* always use only one arg */
		char *e;
		for (i=0; i<6; ++i) {
			bd->bi_enetaddr[i] = s ? simple_strtoul(s, &e, 16) : 0;
			if (s) s = (*e) ? e+1 : e;
		}
	} else if (strcmp(argv[1],"ipaddr") == 0) {
		char *s = argv[2];	/* always use only one arg */
		char *e;
		bd->bi_ip_addr = 0;
		for (i=0; i<4; ++i) {
			ulong val = s ? simple_strtoul(s, &e, 10) : 0;
			bd->bi_ip_addr <<= 8;
			bd->bi_ip_addr  |= (val & 0xFF);
			if (s) s = (*e) ? e+1 : e;
		}
	}
}

/* need both ENV and flash */
#if ((CONFIG_COMMANDS & (CFG_CMD_ENV|CFG_CMD_FLASH)) == (CFG_CMD_ENV|CFG_CMD_FLASH))

void do_saveenv  (cmd_tbl_t *cmdtp, bd_t *bd, int flag, int argc, char *argv[])
{
	int rc;
	extern void flash_sect_protect (int p, ulong addr_first, ulong addr_last);
	extern void flash_sect_erase (ulong addr_first, ulong addr_last);
#ifdef CONFIG_4xx
	/*
	 * On ppc4xx still saved somewhere within a flash sector (no sector
	 * reserved for the environment variables). This will be changed in a
	 * future release.
	 */
        ulong sector_flash_addr;
        ulong sector_flash_size;
        ulong sector_flash_offs;
        int i;
        flash_info_t *info;

	env_init();

        /*
         * Calculate environment variables sector address and size
         */
        info = addr2info((ulong)flash_addr);
        for (i=0; i<info->sector_count; i++)
          {
            if (info->start[i] >= (ulong)flash_addr)
              break;
          }
        sector_flash_addr = info->start[i-1];
        sector_flash_size = info->start[i] - info->start[i-1];
        sector_flash_offs = (ulong)flash_addr - info->start[i-1];

        /*
         * Copy sector down to ram
         */
	memcpy((uchar *)0x10000, (uchar *)sector_flash_addr, sector_flash_size);

        /*
         * Copy new environment variables to ram image of flash sector
         */
	memcpy((uchar *)(0x10000+sector_flash_offs), (uchar *)environment, env_size);

	flash_sect_protect (0, sector_flash_addr, sector_flash_addr+sector_flash_size-1);

	printf ("Erasing Flash...");
	flash_sect_erase (sector_flash_addr, sector_flash_addr+sector_flash_size-1);

	printf ("Saving Environment to Flash...\n");
	switch (rc = flash_write ((uchar *)0x10000, sector_flash_addr, sector_flash_size)) {
	case 0: break;
	case 1: printf ("Timeout writing to Flash\n");
		break;
	case 2: printf ("Flash not Erased\n");
		break;
	case 4: printf ("Can't write to protected Flash sectors\n");
		break;
	default:
		printf ("%s[%d] FIXME: rc=%d\n",__FILE__,__LINE__,rc);
	}
	
	flash_sect_protect (1, sector_flash_addr, sector_flash_addr+sector_flash_size-1);
#else	/* ! CONFIG_CPCI405 */

	env_init();

	flash_sect_protect (0, (ulong)flash_addr, (ulong)flash_addr+env_size-1);
	
	printf ("Erasing Flash...");
	flash_sect_erase ((ulong)flash_addr, (ulong)flash_addr+env_size-1);
	
	printf ("Saving Environment to Flash...\n");
	switch (rc = flash_write (environment, (ulong)flash_addr, env_size)) {
	case 0: break;
	case 1: printf ("Timeout writing to Flash\n");
		break;
	case 2: printf ("Flash not Erased\n");
		break;
	case 4: printf ("Can't write to protected Flash sectors\n");
		break;
	case 8: printf ("Outside available Flash\n");
		return;
	default:
		printf ("%s[%d] FIXME: rc=%d\n",__FILE__,__LINE__,rc);
	}
	
	flash_sect_protect (1, (ulong)flash_addr, (ulong)flash_addr+env_size-1);

#endif	/* CONFIG_CPCI405 */
}

#endif	/* CFG_CMD_ENV + CFG_CMD_FLASH */

/*
 * s1 is either a simple 'name', or a 'name=value' pair.
 * s2 is a 'name=value' pair.
 * If the names match, return the value of s2, else NULL.
 */

static uchar *
envmatch (uchar *s1, uchar *s2)
{

	while (*s1 == *s2++)
		if (*s1++ == '=')
			return(s2);
	if (*s1 == '\0' && *(s2-1) == '=')
		return(s2);
	return(NULL);
}

/*
 * Prevent confusion if running from erased flash memory
 */
static void env_init(void)
{
	if (environment[0] == 0xFF) {
		printf ("*** Warning - no Environment, using defaults\n\n");
		memcpy (environment,
			default_environment,
			sizeof(default_environment));
	}
}
